
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>

#include "../../src/initng_global.h"
#include "../../src/process_db.h"
#include "../../src/service_db.h"
#include "../../src/initng_handler.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include <utmp.h>
#include "initreq.h"

int utmp_stored = FALSE;

/* static local function here */
static void parse_control_input(void);
static void initctl_control_close(void);
static int initctl_control_open(void);

int pipe_fd = -1;               /* /dev/initctl */
struct stat st, st2;

#define PIPE_FD    10                       /* Fileno of initfifo. */

static void initctl_control_close(void)
{
    if (pipe_fd >= 0)
    {
        initng_del_hook(FDWATCHERS, &parse_control_input);
        close(pipe_fd);
        pipe_fd = -1;
    }
}

static int initctl_control_open(void)
{
    D_("initctl control open (%d)\n", pipe_fd);
    /* First, try to create /dev/initctl if not present. */
    if (stat(INIT_FIFO, &st2) < 0 && errno == ENOENT)
        (void) mkfifo(INIT_FIFO, 0600);

    /*
     *  If /dev/initctl is open, stat the file to see if it
     *  is still the _same_ inode.
     */
    if (pipe_fd >= 0)
    {
        fstat(pipe_fd, &st);
        if (stat(INIT_FIFO, &st2) < 0 ||
            st.st_dev != st2.st_dev || st.st_ino != st2.st_ino)
        {
            initctl_control_close();
        }
    }

    /* If the pipe isn't open, try to open it. */
    if (pipe_fd <= 0)
    {
        if ((pipe_fd = open(INIT_FIFO, O_RDWR | O_NONBLOCK)) >= 0)
        {
            D_("Opened on fd %i\n", pipe_fd);
            fstat(pipe_fd, &st);
            if (!S_ISFIFO(st.st_mode))
            {
                /* /dev/initctl is there, but we can't open it */
                F_("%s is not a fifo\n", INIT_FIFO);
                if (pipe_fd >= 0)
                {
                    close(pipe_fd);
                    pipe_fd = -1;
                }
                return (FALSE);
            }
        }
        if ((pipe_fd >= 0) && (pipe_fd < PIPE_FD))
        {
            /*
             *  Don't use fd's 0, 1 or 2.
             */
            (void) dup2(pipe_fd, PIPE_FD);
            if (pipe_fd >= 0)
                close(pipe_fd);
            pipe_fd = PIPE_FD;
        }
        initng_add_hook(FDWATCHERS, &parse_control_input, pipe_fd);
    }
    return (TRUE);
}

/* To be called when there is input on the control bus */
void parse_control_input(void)
{
    int n;
    struct init_request request;

    /* Read data from /dev/initctl */
    n = read(pipe_fd, &request, sizeof(request));

    /* Check if request is ok : */
    if (n == 0)
    {
        F_("read 0 bytes, this should never happen!\n");
        return;
    }
    if (n <= 0)
    {
        if (errno == EINTR)
            return;
        F_("Error reading request\n");
        return;
    }
    if (request.magic != INIT_MAGIC || n != sizeof(request))
    {
        F_("got bogus initrequest\n");
        return;
    }

    /* Todo, handle these:
       #define INIT_CMD_START              0
       #define INIT_CMD_RUNLVL        1
       #define INIT_CMD_POWERFAIL     2
       #define INIT_CMD_POWERFAILNOW  3
       #define INIT_CMD_POWEROK       4
     */

    /* Request is OK, handle it: */
    D_("init data is : - %d (%c),%d (%c).\n", request.runlevel,
       request.runlevel, request.cmd, request.cmd);
    if (request.cmd == INIT_CMD_RUNLVL)
        if ((request.runlevel == '0') || (request.runlevel == '6'))
        {
            D_("Setting new runlevel\n");
            g.when_out = THEN_NEW_RUNLEVEL;
            if (request.runlevel == '0')
            {
                D_("Setting HALT\n");
                s_set_string(S_RUNLEVEL, strdup("halt"));
            }
            else
            {
                D_("Setting REBOOT\n");
                s_set_string(S_RUNLEVEL, strdup("reboot"));
            }
            D_("Stopping all processes\n");
            stop_all();
            return;
        }
    D_("got unimplemented initrequest - %d (%c),%d (%c).\n", request.runlevel,
       request.runlevel, request.cmd, request.cmd);
}

int makeutmp(int runlevel)
{
    D_("Making utmp file for runlevel %d\n", runlevel);
    struct utmp utmp;
    time_t t;

    /*
     * Dont try if UTMP_FILE isn't there, maybe we should force create this if it isn't there? - todo, check that.
     */
    if (access(UTMP_FILE, F_OK) < 0)
        return (FALSE);
    /*
       todo, is this a good idea or a bad idea?
       utmpname("/var/run/utmp");
     */

    setutent();
    memset(&utmp, 0, sizeof(utmp));
    utmp.ut_type = RUN_LVL;
    utmp.ut_pid = ('#' << 8) + runlevel + '0';
    time(&t);
    utmp.ut_time = (int) t;
    if (pututline(&utmp) == NULL)
    {
        D_("pututline failed\n");
        endutent();
        return (FALSE);
    }
    endutent();
    return (TRUE);
}

void service_status(active_h * service)
{
    if (service->a_status == LOADING
        || service->a_status == WAITING_FOR_START_DEP
        || service->a_status == START_DEP_MET)
        return;
    if (!initctl_control_open())
    {
        F_("Warning, failed to open /dev/initctl\n");
        return;
    }
    if (!utmp_stored)
        if (makeutmp(3))
            utmp_stored = TRUE;
}

void is_system_up(h_sys_state state)
{
    if ((state == STATE_UP) || (state == STATE_STOPING))
    {
        initng_del_hook(SWATCHERS, &is_system_up);
        initng_del_hook(WATCHERS, &service_status);
    }
}

int module_init(void)
{
    if (!s_is(I_AM_INIT))
        return (TRUE);                      /* this is not a failure */
    /* return (FALSE); */

    utmp_stored = FALSE;

    D_("moule_init();\n");
    initctl_control_open();

    initng_add_hook(SWATCHERS, &is_system_up);

    if (!initng_add_hook(WATCHERS, &service_status))
    {
        F_("Fail add hook!\n");
        return (FALSE);
    }
    return (TRUE);
}


void module_unload(void)
{
    D_("moule_unload();\n");
    initctl_control_close();
    /* remove all hooks */
    if (!initng_del_hook(WATCHERS, &service_status))
    {
        F_("initctl.so.0.0 failed unloding watcher!\n");
    }
}
