
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#include "print_service.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "../../src/initng_global.h"
#include "../../src/process_db.h"
#include "../../src/service_db.h"
#include "../../src/initng_handler.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_global.h"

#include "initng_ngc.h"

/* static local function here */
static void parse_control_input(void);
static void remote_control_close(void);
static int remote_control_open(void);
void fail(void);                /* Avoid using goto */

int infd, outfd;

#define MIN_FD 11                           /* Minimum FD */


static void remote_control_close(void)
{
    if (infd)
        close(infd);
    if (outfd)
        close(outfd);
    infd = 0;
    outfd = 0;
}

int checkfd(int fd, int min)
{
    if ((fd >= 0) && (fd < min))
    {
        /*
         *  Don't use fd's 0, 1 or 2.
         */
        (void) dup2(fd, min);
        if (fd >= 0)
            close(fd);
        fd = min;
    }
    return fd;
}

static int remote_control_open(void)
{
    struct stat st;

    /* Open up control chanel */
    if (s_is(I_AM_INIT))
    {
        /* try creat fifos is missing */
        if (stat(CONTROL_PATH "/in", &st) < 0 && errno == ENOENT)
            if (mkfifo(CONTROL_PATH "/in", 0600) < 0)
            {
                D_("System fifo path not writable\n");  /* do a quite debug warning */
                return (FALSE);
            }
        if (stat(CONTROL_PATH "/out", &st) < 0 && errno == ENOENT)
            if (mkfifo(CONTROL_PATH "/out", 0600) < 0)
            {
                D_("System fifo path not writable\n");  /* do a quite debug warning */
                return (FALSE);
            }


        infd = open(CONTROL_PATH "/in", O_RDWR);
        outfd = open(CONTROL_PATH "/out", O_RDWR | O_NONBLOCK);
    }
    else
    {
        /* open a testing control port, that ngcd can comunicate with */
        mkfifo(CONTROL_PATH "/in-test", 0600);
        mkfifo(CONTROL_PATH "/out-test", 0600);
        infd = open(CONTROL_PATH "/in-test", O_RDWR);
        outfd = open(CONTROL_PATH "/out-test", O_RDWR | O_NONBLOCK);
    }

    /* Check if channel out and in is clear */
    if (infd < 0 || outfd < 0)
    {
        fail();
        return (FALSE);
    }

    infd = checkfd(infd, MIN_FD);
    outfd = checkfd(outfd, MIN_FD + 1);

    if (fstat(infd, &st) < 0 || !S_ISFIFO(st.st_mode) ||
        fstat(outfd, &st) < 0 || !S_ISFIFO(st.st_mode))
    {
        F_("Not a fifo file!\n");
        fail();
        return (FALSE);
    }

    /* set close-on-exec, what does this mean in practise? */
    fcntl(infd, F_SETFD, FD_CLOEXEC);
    fcntl(outfd, F_SETFD, FD_CLOEXEC);

    return (TRUE);
}

void fail(void)
{
    if (s_is(I_AM_INIT))
    {
        F_("error open " CONTROL_PATH "/in and " CONTROL_PATH "/out\n");
    }
    else
    {
        F_("error open " CONTROL_PATH "/in-test and "
           CONTROL_PATH "/out-test\n");
    }
}

/* To be called when there is input on the control bus */
void parse_control_input(void)
{
    char c_input[101];
    active_h *apt = NULL;       /* a temp active struct */

    /* to be sure null end while reading */
    c_input[read(infd, c_input, 100)] = '\0';

    D_("parse_control_input(): INPUT IS: \"%s\"\n", c_input);

    /* first char in input is a command key, switching it */
    switch (c_input[0])
    {

        case 'g':                          /* get pid of */
            {
                char pid[101];

                if (c_input + 1 &&
                    (apt = active_db_find_by_name(c_input + 1)) &&
                    apt && apt->start_process && apt->start_process->pid)
                {
                    sprintf(pid, "%i", apt->start_process->pid);
                    write(outfd, pid, strlen(pid));
                }
                else
                {
                    write(outfd, "0", 1);
                }
                break;
            }

        case 'u':                          /* start service */
            {
                if (c_input + 1 && start_new_service_named(c_input + 1))
                {
                    write(outfd, "1", 1);
                }
                else
                {
                    write(outfd, "0", 1);
                }
                break;
            }

        case 'j':                          /* stop_all then start */
            {
                if (c_input + 1)
                {
                    g.when_out = THEN_NEW_RUNLEVEL;
                    s_set_string(S_RUNLEVEL, i_strdup(c_input + 1));
                    stop_all();
                    write(outfd, "1", 1);
                }
                else
                {
                    write(outfd, "0", 1);
                }
                break;
            }

        case 'z':                          /* free service */
            {
                if (c_input + 1)
                {
                    apt = active_db_find_by_name(c_input + 1);
                    active_db_free(apt);
                    write(outfd, "1", 1);
                }
                else
                {
                    write(outfd, "0", 1);
                }
                break;
            }

        case 'd':                          /* take down service */
            if (c_input + 1 &&
                (apt = active_db_find_by_name(c_input + 1))
                && (stop_service(apt)))
                write(outfd, "1", 1);
            else
                write(outfd, "0", 1);
            break;

        case 'h':
            {                               /* print history */
                char send_buf[2001];
                history_h *tmp = NULL;
                int max = 0;

                D_("parse_control_input(): Sending History\n history:");
                while ((tmp = history_get_next(tmp)))
                {
                    if (tmp && tmp->name)
                    {
                        D_(" %s ", tmp->name);
                        sprintf(send_buf, "%i|%li|%s\n", tmp->action,
                                (long int) tmp->time, tmp->name);
                        send_buf[2001] = '\0';  /* to make sure it stops on buffer overflow */
                        write(outfd, send_buf, strlen(send_buf));
                    }
                    if (max++ >= (HISTORY - 1))
                        break;
                }
            }
            break;

        case 's':
            {                               /* print active list */
                char send_buf[2001];

                D_("parse_control_input(): Sending Running process list\n sending :");
                while ((apt = active_db_next(apt)))
                {
                    if (apt && apt->name && apt->a_status)
                    {
                        D_(" %s->%i ", apt->name, apt->a_status);
                        sprintf(send_buf, "%i|%li|%s\n", apt->a_status,
                                (long int) apt->time_got_status, apt->name);
                        write(outfd, send_buf, strlen(send_buf));
                    }
                }
            }
            break;

        case 't':                          /* print uptime */
            if (c_input + 1 && (apt = active_db_find_by_name(c_input + 1)))
            {
                char tmp_time[10];

                sprintf(tmp_time, "%i",
                        (int) difftime(time(0), apt->time_got_status));
                write(outfd, tmp_time, strlen(tmp_time));
            }
            break;

        case 'r':
            {                               /* reload service data, actually removes service data, and it will be loaded when needed. */
                service_h *s;

                if (c_input + 1 && (s = service_db_find_by_name(c_input + 1)))
                {
                    D_("removing service data for %s!\n", c_input + 1);
                    service_db_free(s);
                    write(outfd, "1", 1);
                }
                else if (c_input + 1
                         && strncasecmp(c_input + 1, "all", 3) == 0)
                {
                    D_("Clearing complete db.\n");
                    service_db_clear_all();
                    write(outfd, "1", 1);
                }
                else
                {
                    F_("did not find service %s to clear.\n", c_input + 1);
                    write(outfd, "0", 1);
                }
            }
            break;

        case 'x':                          /* restart initng */
            write(outfd, "1", 1);           /* write status before, because it can't after */
            g.when_out = THEN_RESTART;
            stop_all();
            break;
        case 'i':                          /* close /in and /out sockets */
            write(outfd, "1", 1);           /* do this before closing */
            remote_control_close();
            break;
#ifdef RUNTIME_DEBUG
        case 'v':                          /* close /in and /out sockets */
            if (g.verbose == 0)
            {
                g.verbose = 1;
                write(outfd, "2", 1);
            }
            else
            {
                g.verbose = 0;
                write(outfd, "1", 1);
            }
            break;
#endif
        case 'q':                          /* quit initng */
            write(outfd, "1", 1);           /* write status before, because it can't after */
            g.when_out = THEN_QUIT;
            stop_all();
            break;
#ifdef PRINT_SERVICE
        case 'p':                          /* print special service info or all to the initng console */
            {
                service_h *s;

                if (c_input + 1 && (s = service_db_find_by_name(c_input + 1)))
                {
                    service_db_print(s);
                }
                else
                {
                    service_db_print_all();
                }
                write(outfd, "1", 1);
            }
            break;
#endif
        case 'n':                          /* send percent done */
            {
                char send[10];

                sprintf(send, "%i", active_db_percent_started());
                write(outfd, send, 4);
            }
            break;
        case 'P':
            {                               /* set PID to in process_db */
                char *sname = NULL;
                int pid = 0;
                sname = initng_calloc(100, sizeof(char));
                sscanf(c_input, "P%i:%s", &pid, sname);
                if (pid && sname)
                {
                    apt = active_db_find_by_name(sname);
                    if (apt && apt->name && apt->start_process)
                    {
                        D_("setting pid of %s from %i to %i!\n",
                           apt->name, apt->start_process->pid, pid);
                        apt->start_process->pid = pid;
                        if (active_db_is(UP_WHEN_PID_SET, apt))
                            apt->a_status = RUNNING;
                        write(outfd, "1", 1);
                    }
                    else
                    {
                        F_("did not found service %s to set pid %i to.",
                           sname, pid);
                        write(outfd, "0", 1);
                    }
                }
                else
                {
                    F_("Bad controlcall trying to change pid\n");
                    write(outfd, "0", 1);
                }
                break;
            }
        default:
            F_("parse_control_input(): INVALID CONTROL REQUEST\n");
    }
    write(outfd, "\0", 1);
}

void check_initial_up(active_h * service)
{
    if (!service)
        return;
    /* open remote control fifo */
    if (!remote_control_open())
        return;
    /* add us to list of parsers */
    initng_add_hook(FDWATCHERS, &parse_control_input, infd);
    initng_del_hook(WATCHERS, &check_initial_up);
}

int module_init(void)
{
    D_("moule_init();\n");
    initng_add_hook(WATCHERS, &check_initial_up);
    return (TRUE);
}


void module_unload(void)
{
    remote_control_close();

    D_("moule_unload();\n");

    initng_del_hook(FDWATCHERS, &parse_control_input);
    initng_del_hook(WATCHERS, &check_initial_up);
}
