/*
 * splash.c - The core of splash_util
 *
 * Copyright (C) 2004, Michal Januszewski <spock@gentoo.org>
 * 
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file COPYING in the main directory of this archive for
 * more details.
 *
 * $Header: /srv/cvs/splash/utils/splash.c,v 1.17 2005/01/16 09:39:58 spock Exp $ 
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <linux/fb.h>
#include <linux/kd.h>
#include <linux/vt.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <unistd.h>

#include "splash.h"
#include "libsplash.h"

void set_progress(char *theme, int progress)
{
    /* check if mode is up */
    DEBUG("set_progress(%s,%i);\n",theme, progress);
    if(get_mode()) {
	DEBUG("set_progress(): Not in silent mode\n");
	return;
    }
    /* set defaults */
    arg_vc = 0;
    arg_mode = 's';
    arg_task = paint;

    /* set some default globals */
    arg_theme = theme;
    arg_progress = progress;

    /* open framebuffer */
    get_fb_settings(arg_fb);

    /* do som theme parsing */
    config_file = get_cfg_file(arg_theme);
    parse_cfg(config_file);

    /* draw progressbar */
    if (do_getpic(FB_SPLASH_IO_ORIG_USER, 0, arg_mode))
    {
        return;
    }

    /* free on return */
    free((u8 *) pic.data);
    if (pic.cmap.red)
        free(pic.cmap.red);
}

/* is some sort of splash verbose or silent enabled ? */
int get_state()
{
    unsigned int i;
    int fp;
    struct fb_splash_iowrapper wrapper = {
        .vc = arg_vc,
        .origin = FB_SPLASH_IO_ORIG_USER,
        .data = &i,
    };

    fp = open(SPLASH_DEV, O_WRONLY);
    if (fp == -1)
    {
        DEBUG("Can't open %s\n", SPLASH_DEV);
        return (-1);
    }
    ioctl(fp, FBIOSPLASH_GETSTATE, &wrapper);
    close(fp);

    DEBUG("Splash state on console %d: %s\n", arg_vc,
           (i != 0) ? "on" : "off");
    return (i);
}

int get_mode()
{
    unsigned int i;
    int fp;
    struct vt_stat stat;

    fp = open(SPLASH_DEV, O_WRONLY);
    if (fp == -1)
    {
        DEBUG("Can't open %s\n", SPLASH_DEV);
        return (-1);
    }

    if (ioctl(fp, VT_GETSTATE, &stat) != -1)
      if(stat.v_active == TTY_SILENT)
	i = 1;
    close(fp);

    DEBUG("Splash mode: %s\n", (i) ? "silent" : "verbose");
    return (i);
}




#ifdef COMPILE_MAIN
int main(int argc, char **argv)
{
    char dev[16];
    unsigned int c, i, y;
    int fp, err = 0;
    struct vt_stat stat;

    detect_endianess();
    arg_task = none;

    while ((c = getopt_long(argc, argv, "c:t:m:p:h", options, NULL)) != EOF)
    {

        switch (c)
        {

            case 'h':
                usage();
                return 0;

            case 0x100:
                arg_fb = atoi(optarg);
                break;
            case 0x101:
                arg_vc = atoi(optarg);
                break;

            case 0x103:
            case 't':
                arg_theme = optarg;
                break;

            case 0x102:
            case 'c':
                for (i = 0; i < sizeof(cmds) / sizeof(struct cmd); i++)
                {

                    if (!strcmp(optarg, cmds[i].name))
                    {
                        arg_task = cmds[i].value;
                        break;
                    }
                }
                break;

            case 'm':
            case 0x104:
                if (optarg[0] == 's')
                    arg_mode = 's';
                else
                    arg_mode = 'v';
                break;

            case 'p':
            case 0x105:
                arg_progress = atoi(optarg);
                break;
        }
    }

    if (arg_task == none)
    {
        usage();
        return 0;
    }

    get_fb_settings(arg_fb);

    if (arg_theme)
        config_file = get_cfg_file(arg_theme);

    if (config_file)
        parse_cfg(config_file);

    /* we've got to repaint the whole screen if we have icons to draw */
    /*
       if (arg_task == paint && cf_icons_cnt > 0)
       arg_task = repaint;
     */

    switch (arg_task)
    {

        case on:
            cmd_setstate(1, FB_SPLASH_IO_ORIG_USER);
            break;

        case off:
            cmd_setstate(0, FB_SPLASH_IO_ORIG_USER);
            break;

        case setpic:
            {
                struct vt_stat stat;

                if ((fp = open("/dev/tty", O_NOCTTY)) != -1)
                {
                    if (ioctl(fp, VT_GETSTATE, &stat) != -1)
                    {
                        if (arg_vc != stat.v_active - 1)
                            goto setpic_out;
                    }
                    close(fp);
                }

                do_getpic(FB_SPLASH_IO_ORIG_USER, 1, arg_mode);
              setpic_out:break;
            }

        case setcfg:
            do_config(FB_SPLASH_IO_ORIG_USER);
            break;

        case getstate:
            {
                struct fb_splash_iowrapper wrapper = {
                    .vc = arg_vc,
                    .origin = FB_SPLASH_IO_ORIG_USER,
                    .data = &i,
                };

                fp = open(SPLASH_DEV, O_WRONLY);
                if (fp == -1)
                {
                    fprintf(stderr, "Can't open %s\n", SPLASH_DEV);
                    break;
                }
                ioctl(fp, FBIOSPLASH_GETSTATE, &wrapper);
                close(fp);

                printf("Splash state on console %d: %s\n", arg_vc,
                       (i != 0) ? "on" : "off");
                break;
            }

        case setmode:
            fp = open(SPLASH_DEV, O_WRONLY);
            if (fp == -1)
            {
                fprintf(stderr, "Can't open %s\n", SPLASH_DEV);
                break;
            }
            ioctl(fp, FBIOSPLASH_SETMODE,
                  (arg_mode ==
                   's') ? FB_SPLASH_MODE_SILENT : FB_SPLASH_MODE_VERBOSE);
            close(fp);

            if (arg_mode == 's')
            {
                fp = open("/dev/vc/0", O_WRONLY);
                if (fp == -1)
                {
                    fprintf(stderr, "Can't open /dev/vc/0\n");
                    break;
                }
                ioctl(fp, KDSETMODE, KD_GRAPHICS);
                close(fp);
            }

            break;

        case getmode:
            fp = open(SPLASH_DEV, O_WRONLY);
            if (fp == -1)
            {
                fprintf(stderr, "Can't open %s\n", SPLASH_DEV);
                break;
            }
	    if (ioctl(fp, VT_GETSTATE, &stat) != -1)
	      if(stat.v_active == TTY_SILENT)
		i = 1;
	    close(fp);

            printf("Splash mode: %s\n", (i) ? "silent" : "verbose");
            break;

        case getcfg:
            cmd_getcfg(FB_SPLASH_IO_ORIG_USER);
            break;

        case paint:
        case repaint:
            if (do_getpic(FB_SPLASH_IO_ORIG_USER, 0, arg_mode))
            {
                err = -1;
                break;
            }

            if (arg_progress > 0 && arg_task == paint && !cf_rects_cnt)
                goto paint_out;

            sprintf(dev, "/dev/fb%d", arg_fb);
            if ((c = open(dev, O_WRONLY)) == -1)
            {
                sprintf(dev, "/dev/fb/%d", arg_fb);
                if ((c = open(dev, O_WRONLY)) == -1)
                {
                    printerr("Failed to open /dev/fb%d or /dev/fb/%d.\n",
                             arg_fb, arg_fb);
                    break;
                }
            }

            if (pic.cmap.red)
                ioctl(c, FBIOPUTCMAP, &pic.cmap);

            if (arg_task == repaint)
            {

                i = fb_var.xres * ((fb_var.bits_per_pixel + 7) >> 3);

                for (y = 0; y < fb_var.yres; y++)
                {
                    if (i != fb_fix.line_length || fb_var.yoffset != 0)
                    {
                        lseek(c, (fb_var.yoffset + y) * fb_fix.line_length,
                              SEEK_SET);
                    }

                    write(c, pic.data + i * y, i);
                }

            }
            else
            {
                int j;
                int bytespp = ((fb_var.bits_per_pixel + 7) >> 3);

                for (i = 0; i < cf_rects_cnt; i++)
                {
                    j = (cf_rects[i].x2 - cf_rects[i].x1) * bytespp;

                    for (y = cf_rects[i].y1; y < cf_rects[i].y2; y++)
                    {
                        lseek(c,
                              (fb_var.yoffset + y) * fb_fix.line_length +
                              cf_rects[i].x1 * bytespp, SEEK_SET);
                        write(c,
                              pic.data + (y * fb_var.xres +
                                          cf_rects[i].x1) * bytespp, j);
                    }
                }
            }

            close(c);

          paint_out:free((u8 *) pic.data);
            if (pic.cmap.red)
                free(pic.cmap.red);
            if (do_getpic(FB_SPLASH_IO_ORIG_USER, 0, arg_mode))
            {
                err = -1;
                break;
            }

            if (arg_progress > 0 && arg_task == paint && !cf_rects_cnt)
                goto paint_out;

            sprintf(dev, "/dev/fb%d", arg_fb);
            if ((c = open(dev, O_WRONLY)) == -1)
            {
                sprintf(dev, "/dev/fb/%d", arg_fb);
                if ((c = open(dev, O_WRONLY)) == -1)
                {
                    printerr("Failed to open /dev/fb%d or /dev/fb/%d.\n",
                             arg_fb, arg_fb);
                    break;
                }
            }

            if (pic.cmap.red)
                ioctl(c, FBIOPUTCMAP, &pic.cmap);

            if (arg_task == repaint)
            {

                i = fb_var.xres * ((fb_var.bits_per_pixel + 7) >> 3);

                for (y = 0; y < fb_var.yres; y++)
                {
                    if (i != fb_fix.line_length || fb_var.yoffset != 0)
                    {
                        lseek(c, (fb_var.yoffset + y) * fb_fix.line_length,
                              SEEK_SET);
                    }

                    write(c, pic.data + i * y, i);
                }

            }
            else
            {
                int j;
                int bytespp = ((fb_var.bits_per_pixel + 7) >> 3);

                for (i = 0; i < cf_rects_cnt; i++)
                {
                    j = (cf_rects[i].x2 - cf_rects[i].x1) * bytespp;

                    for (y = cf_rects[i].y1; y < cf_rects[i].y2; y++)
                    {
                        lseek(c,
                              (fb_var.yoffset + y) * fb_fix.line_length +
                              cf_rects[i].x1 * bytespp, SEEK_SET);
                        write(c,
                              pic.data + (y * fb_var.xres +
                                          cf_rects[i].x1) * bytespp, j);
                    }
                }
            }

            close(c);

          paint_out:free((u8 *) pic.data);
            if (pic.cmap.red)
                free(pic.cmap.red);
            break;

        default:
            break;
    }

    if (config_file)
        free(config_file);

    return err;
}
#endif
